if __name__ == '__main__':
    print('Tony has been updated and this way of running the program is deprecated. Please refer to the README for instructions: https://github.com/Pasu4/neuro-api-tony?tab=readme-ov-file#installation')
    input('\nPress Enter to exit...')